#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
using std::cout;
using std::endl;

#include "kino_common.h"
#include "page.h"

#include "page_bttv.h"
#include "displayer.h"
#include "v4l.h"

extern KinoCommon *common;

extern "C" {
#include "support.h"


void
on_v4l_capture_page_record_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{

	common->getPageBttv()->startCapture();
}


void
on_v4l_capture_page_stop_button_clicked
                                        (GtkButton       *button,
                                        gpointer         user_data)
{
	common->getPageBttv()->stopCapture();
}

}

PageBttv::PageBttv( KinoCommon *common ) : seeking( false ) { 
	this->common = common;
}

PageBttv::~PageBttv() {
}

void PageBttv::start() {
	cout << ">>> Starting bttv" << endl;
	v4l = new GDKV4L( lookup_widget( common->getWidget(), "drawingarea3" ) );
	GtkEntry *fileEntry = GTK_ENTRY( lookup_widget(common->getWidget(), "entry1") );
	if ( !strcmp( gtk_entry_get_text( fileEntry ), "" ) )
		gtk_entry_set_text( fileEntry, Preferences::getInstance().file );

	Preferences &prefs = Preferences::getInstance();
	v4l->setInfo( prefs.v4lVideoDevice, prefs.v4lInput, prefs.v4lAudioDevice, atoi( prefs.v4lAudio ) );
	
	if ( v4l->openDevice() ) {
		v4l->report();
		v4l->setChannel( 0 );
		v4l->setTuner( 0 );
		v4l->startAudio();
		v4l->startVideo();
	}
	else {
		cout << "unable to open video device" << endl;
	}
}

gulong PageBttv::activate() {
	if ( v4l->deviceAvailable() ) 
		return VIDEO_START_OF_SCENE | VIDEO_REWIND | VIDEO_BACK | 
		   	   VIDEO_PLAY | VIDEO_STOP | 
			   VIDEO_FAST_FORWARD | VIDEO_FORWARD | VIDEO_NEXT_SCENE;
	else
		return 0;
}

void PageBttv::clean() {
	if ( v4l->deviceAvailable() ) {
		v4l->stopVideo();
		v4l->stopAudio();
	}
	delete v4l;
	v4l = NULL;
}

void PageBttv::StartOfMovie() {
	v4l->setFrequency( 0 );
}

void PageBttv::videoPreviousScene() {
	v4l->setFrequency( v4l->getFrequency() - 100 );
}

void PageBttv::videoRewind() {
	if ( seeking == true )
		return;
	seeking = true;
	do {
		if ( ( v4l->getFrequency() - 1 ) > 0 )
			v4l->setFrequency( v4l->getFrequency() - 1 );
		else
			break;
	}
	while( v4l->getSignal() != 0 );
	do {
		if ( ( v4l->getFrequency() - 2 ) > 0 )
			v4l->setFrequency( v4l->getFrequency() - 2 );
		else
			break;
		cout << "frequency = " << v4l->getFrequency() << endl;
		while (gtk_events_pending()) {
        	gtk_main_iteration();
		}
	}
	while( v4l->getSignal() == 0 && seeking );
	seeking = false;
}

void PageBttv::videoPlay() {
	v4l->startVideo();
}

void PageBttv::videoStop() {
	seeking = false;
	v4l->stopVideo();
}

void PageBttv::videoFastForward() {
	if ( seeking == true )
		return;
	seeking = true;
	do {
		if ( ( v4l->getFrequency() + 1 ) < 0xffff )
			v4l->setFrequency( v4l->getFrequency() + 1 );
		else
			break;
	}
	while( v4l->getSignal() != 0 );
	do {
		if ( ( v4l->getFrequency() + 2 ) < 0xffff )
			v4l->setFrequency( v4l->getFrequency() + 2 );
		else
			break;
		cout << "frequency = " << v4l->getFrequency() << endl;
		while (gtk_events_pending()) {
        	gtk_main_iteration();
		}
	}
	while( v4l->getSignal() == 0 && seeking );
	seeking = false;
}

void PageBttv::videoBack() {
	v4l->setFrequency( v4l->getFrequency() - 1 );
}

void PageBttv::videoForward() {
	v4l->setFrequency( v4l->getFrequency() + 1 );
}

void PageBttv::videoNextScene() {
	v4l->setFrequency( v4l->getFrequency() + 100 );
}

void PageBttv::startCapture() {
	v4l->startCapturing();
}

void PageBttv::stopCapture() {
	v4l->stopCapturing();
}

void PageBttv::saveFrame() {
	cout << "BTTV Save Frame Requested" << endl;
}

void PageBttv::view50percent() {
	cout << "BTTV View 50% Requested" << endl;
}

void PageBttv::view100percent() {
	cout << "BTTV View 100% Requested" << endl;
}



