/*
 * page_export_mjpeg.cc -- Notebook Firewire/AVI/Still Frame Export Page Object
 * Copyright (C) 2002 Charles Yates <charles.yates@pandora.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "kino_av_pipe.h"

#include <vector>
#include <iostream>
using std::cout;
using std::endl;

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>

#include "page_export_mjpeg.h"
#include "preferences.h"
#include "kino_common.h"
#include "frame.h"
#include "page_editor.h"
#include "message.h"

extern "C" {
#include "support.h"
#include "callbacks.h"
#include "interface.h"
#include "commands.h"

#include <string.h>
#include <signal.h>

}

/** Constructor for page.

\param _exportPage page to which this page belongs
\param _common KinoCommon object
*/

ExportMJPEG::ExportMJPEG( PageExport *_exportPage, KinoCommon *_common ) :
  Export(_exportPage, _common)
{
  cout << "> Creating ExportMJPEG Page" << endl;

  /* Get pointers to own controls */
  fileEntry 
    = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( 
							      lookup_widget( common->getWidget(), 
									     "gnome_fileentry_export_mjpeg" ) ) ) );
  
  notebook 
    = GTK_NOTEBOOK( lookup_widget( exportPage->getWidget(), "notebook_mpeg_divx" ) );	

  mjpegFormat 
    = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( exportPage->getWidget(), 
									  "optionmenu_export_mjpeg_format" ) ) ) );
  deinterlaceMenu 
    = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( common->getWidget(), 
									  "optionmenu_export_mjpeg_deinterlace") ) ) );
  mjpegVideo 
    = GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				"entry_mjpeg_video" ) );	
  mjpegAudio 
    = GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				"entry_mjpeg_audio" ) );	
  mjpegMultiplexer 
    = GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				"entry_mjpeg_multiplexer" ) );	
  splitCheck 
    = GTK_TOGGLE_BUTTON( lookup_widget( common->getWidget(), 
					"checkbutton_export_mjpeg_split" ) );

  divxDeinterlaceMenu 
    = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( common->getWidget(), 
									  "optionmenu_export_divx_deinterlace") ) ) );
  mjpegDivxVideo 
    =  GTK_ENTRY( lookup_widget( exportPage->getWidget(), 
				 "entry_mjpeg_divx" ) );
}

/** Destructor for page.
 */

ExportMJPEG::~ExportMJPEG() {
  cout << "> Destroying ExportMJPEG Page" << endl;
}

/** start exporting MJPEG
 */
void ExportMJPEG::doExport(PlayList * playlist, int begin, int end, int every,
			   bool preview) {
  cout << ">>> ExportMJPEG::startExport" << endl;
  
  static Frame *frame = new Frame;
  // TODO: Fix this buffer stuff.
  char filename[512];

	
  strcpy( filename, gtk_entry_get_text( fileEntry ) );

  if ( !(strcmp(filename, "") && strpbrk( filename, " |`\"" ) == NULL) ) {
    modal_message( "You must enter a filename which doesn't contain special characters (spaces, quotes, pipes etc)." );
    return;
  }
  unsigned char *pixels = new unsigned char[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];

  // Get a sample frame to obtain recording info
  // TODO: Sample frames are problematic... 
  playlist->GetFrame( begin, *frame );
  
  // Get all video and audio info required
  int width = frame->GetWidth();
  int height = frame->GetHeight();
  int isPAL = frame->IsPAL();
  short channels = (short)frame->decoder->audio->num_channels;
  int frequency = frame->decoder->audio->frequency;
  
  // Force audio resampling to that of the first frame
  AudioResample *resampler = new FastAudioResample( frequency );

  // Determine mpeg/divx from selected notebook page
  int format = gtk_notebook_get_current_page( notebook );

  // vars for collecting audio and video pipe commands
  bool split = false;
  kino_video_pipe type = PIPE_VIDEO_MJPEG;
  char audiopipe[ 1024 ] = "";
  char videopipe[ 1024 ] = "";
  char multiplex[ 1024 ] = "";
  char cleanup[ 3 ][ 1024 ] = { "", "", "" };
  
  // MPEG and DIVX translations to commands
  // What started off as a simple addition is getting pretty damn messy :-/ - ho hum
  // Essentially the important things to remember are that only the 5 variables above
  // are populated in the following kludge... (this should move into seprate methods)
  if ( format == 0 ) {
    char *videopipeInput = "";
    int videoformat = 0;
    int deinterlaceRequest = 0;
    char *audioencode = "";
    char *multiplexer = "";
    
    videopipeInput = gtk_entry_get_text( mjpegVideo );
    GtkWidget *active_item = gtk_menu_get_active( mjpegFormat );
    videoformat 
      = g_list_index (GTK_MENU_SHELL (mjpegFormat)->children, active_item);
    audioencode = gtk_entry_get_text( mjpegAudio );
    multiplexer = gtk_entry_get_text( mjpegMultiplexer );
    active_item = gtk_menu_get_active( deinterlaceMenu );
    deinterlaceRequest 
      = g_list_index (GTK_MENU_SHELL (deinterlaceMenu)->children, active_item);
    split = gtk_toggle_button_get_active( splitCheck );
    
    char deinterlace[ 30 ] = "";
    char interlace = '0';
    
    switch( deinterlaceRequest )
      {
      case 0:
	interlace = '1';
	break;
      case 1:
	strcpy( deinterlace, "yuvdenoise -F -f |" );
	break;
      case 2:
	strcpy( deinterlace, "yuvdenoise -F |" );
	break;
      case 3:
	type = PIPE_VIDEO_DEINTERLACED_MJPEG;
	break;
      }
    
    if ( !strstr( videopipeInput, "yuvscaler" ) )
      {
	char scale[ 20 ] = "";
	char bitrate[ 20 ] = "";
	
	switch( videoformat )
	  {
	  case 1:
	    strcpy( scale, "VCD" );
	    break;
	  case 2:
	    strcpy( scale, "VCD" );
	    strcpy( bitrate, "-b 1152" );
	    break;
	  case 4:
	    strcpy( scale, "SVCD" );
	    break;
	  case 5:
	    strcpy( scale, "SVCD" );
	    strcpy( bitrate, "-b 2500" );
	    break;
	  case 6:
	    videoformat = 8;
	    strcpy( scale, "" );
	    break;
	  }
	
	// If the bitrate is specified then remove the defaults above
	if ( strstr( videopipeInput, " -b " ) )
	  strcpy( bitrate, "" );
	
	// Yuck (!) - builds the video pipe depending on data collected above...
	if ( strcmp( scale, "" ) )
	  sprintf( videopipe, 
		   "%s yuvscaler -v 0 -O %s -n %c | %s -f %d -I %c -n %c %s -o %s.tmp",
		   deinterlace, 
		   scale, isPAL ? 'p' : 'n', videopipeInput, videoformat, interlace, isPAL ? 'p' : 'n', bitrate, filename );
	else
	  sprintf( videopipe, "%s %s -f %d -I %c -n %c %s -o %s.tmp", 
		   deinterlace, videopipeInput, videoformat, interlace, 
		   isPAL ? 'p' : 'n', bitrate, filename );
      }
    else
      {
	sprintf( videopipe, "%s %s -f %d -I %c -n %c -o %s.tmp\n",
		 deinterlace, videopipeInput, videoformat, interlace, 
		 isPAL ? 'p' : 'n', filename );
      }
    sprintf( audiopipe, "|%s -o %s.mp2", audioencode, filename );
    sprintf( multiplex, "%s -f %d -o %s%%s.mpeg %s.tmp %s.mp2", 
	     multiplexer, videoformat, filename, filename, filename );
    sprintf( cleanup[ 0 ], "rm -f %s.tmp %s.mp2 %s.wav", 
	     filename, filename, filename );
  }
  else {
    GtkWidget *active_item = gtk_menu_get_active( divxDeinterlaceMenu );
    int deinterlaceRequest = g_list_index (GTK_MENU_SHELL (divxDeinterlaceMenu)->children, active_item);
    char deinterlace[ 30 ] = "";
    char interlace = '0';
    
    switch( deinterlaceRequest )
      {
      case 0:
	interlace = '1';
	break;
      case 1:
	strcpy( deinterlace, "yuvdenoise -F -f |" );
	break;
      case 2:
	strcpy( deinterlace, "yuvdenoise -F |" );
	break;
      case 3:
	type = PIPE_VIDEO_DEINTERLACED_MJPEG;
	break;
      }
    
    sprintf( audiopipe, "%s.wav", filename );
    char *videopipeInput = gtk_entry_get_text( mjpegDivxVideo );
    if ( !strstr( videopipe, "-A" ) )
      sprintf( videopipe, "%s %s -A %s.wav -a %d -o %s.avi", deinterlace, 
	       videopipeInput, filename, frequency == 32000 ? 96 : 128, filename );
    else
      sprintf( videopipe, "%s %s -o %s.avi", deinterlace, videopipeInput, 
	       filename );
    sprintf( cleanup[ 0 ], "rm -f %s.wav", filename );
  }
  
  KinoAudioPipe *wav = KinoAudioFactory::CreateAudioPipe( PIPE_AUDIO_WAV );
  KinoVideoPipe *video = KinoVideoFactory::CreateVideoPipe( type );
  KinoAVPipe *mpeg = new KinoAVPipe( wav, video );

  // Iterate through each scene (opening and closing audio pipes as requested)
  int scene = 0;
  
  for ( int sceneBegin = begin; sceneBegin <= end && exportPage->isExporting; 
	scene ++ ) {
    char full[ 1024 ];
    char sceneString[ 20 ];
    int sceneEnd = end;

    // Determine if we need to split by scene or not
    // when splitting: Output is generated as fileNNN_MMM.mpeg where NNN is the 
    // kino scene number and MMM is the mplex split
    // when not splitting: Output is generated file as fileMMM.mpeg
    if ( split ) {
      sprintf( sceneString, "%03d_%%03d", scene );
      sceneEnd = playlist->FindEndOfScene( sceneBegin ) + 1;
      if ( sceneEnd > end )
	sceneEnd = end;
    }
    else {
      strcpy( sceneString, "%03d" );
    }
    
    mpeg->OpenAudio( audiopipe, channels, frequency, 2 );
    mpeg->OpenVideoPipe( videopipe, width, height );
    
    int imagesize = width * height * 3;
    
    /* Iterate over frames in scene */
    for (int i = sceneBegin; i <= sceneEnd && exportPage->isExporting;
	 i += every) {
      /* Call innerLoopUpdate */
      innerLoopUpdate(i, begin, end, every);
      
      playlist->GetFrame( i, *frame );
      
      resampler->Resample( *frame );
      if ( !mpeg->OutputAudioFrame( resampler->output, resampler->size ) ) {
	modal_message( "Error writing to KINO/MJPEG audio filter - aborting." );
	exportPage->isExporting = false;
      }
      
      if ( exportPage->isExporting ) {
	if ( type == PIPE_VIDEO_MJPEG )
	  frame->ExtractYUV( pixels );
	else
	  frame->ExtractRGB( pixels );
	if ( !mpeg->OutputVideoFrame( pixels, imagesize ) ) {
	  modal_message( "Error writing to KINO/MJPEG video filter - aborting." );
	  exportPage->isExporting = false;
	}
      }
      
      /* Update progressbar ... */
      exportPage->updateProgress( (gfloat) (i + every - begin)  
				  / (gfloat) (end + 1 - begin) );
    }
    
    // Try to finish the stream to create a valid mpeg if possible
    mpeg->CloseAudio();
    mpeg->CloseVideo();
    
    if ( strcmp( multiplex, "" ) ) {
      // expand the template to accomodate the scene splits
      sprintf( full, multiplex, sceneString );
      mpeg->ExecuteCommand( full );
    }
    
    for ( int i = 0; i < 3 && strcmp( cleanup[ i ], "" ); i ++ )
      mpeg->ExecuteCommand( cleanup[ i ] );
    
    // Move to start of next scene
    sceneBegin = sceneEnd + 1;
  }
  delete mpeg;
  delete video;
  delete wav;
  delete resampler;
  
  delete pixels;
}
