/*
 * page_export_stills.cc Notebook Firewire/AVI/Still Frame Export Page Object
 * Copyright (C) 2001 Dan Dennedy <dan@dennedy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <iostream>
using std::cout;
using std::endl;

#include <gtk/gtk.h>

#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>

#include "page_export_stills.h"
#include "preferences.h"
#include "kino_common.h"
#include "frame.h"
#include "page_editor.h"
#include "message.h"

extern "C" {
#include "support.h"
#include "callbacks.h"
#include "interface.h"
#include "commands.h"
}

/** Constructor for page.

\param _exportPage the PageExport object to which this page belongs
\param common	KinoCommon object
*/

ExportStills::ExportStills( PageExport *_exportPage, KinoCommon *_common ) : 
  Export(_exportPage, _common)
{
  cout << "> Creating ExportStills Page" << endl;
	
  /* Get a pointer to the controls of this page */
  qualityScale 
    = GTK_RANGE( lookup_widget( common->getWidget(), "hscale_export_stills") );

  fileEntry 
    = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY(lookup_widget( common->getWidget(), "gnome_fileentry_export_stills" ) ) ) );
}

/** Destructor for page.
 */

ExportStills::~ExportStills() {
  cout << "> Destroying ExportStills Page" << endl;
}

/** start exporting still frames
 */
void ExportStills::doExport(PlayList * playlist, int begin, int end, int every,
			    bool preview) {
  static unsigned char pixels[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];
  static Frame frame;
  GdkImlibSaveInfo saveInfo = {0, 0, 0, 0, 0, 0};
  GdkImlibImage *image = NULL;
  /* TODO: Fix these path handling functions */
  char file[512];
  char filename[512];
  char extension[8];

  strcpy( file, gtk_entry_get_text( fileEntry ) );
  
  /* Check for a valid filename */
  if ( !strcmp(file, "") ) {
    modal_message( "You must enter a filename." );
    return;
  }
  
  /* Take care of the quality parameter */
  GtkAdjustment * adjust = gtk_range_get_adjustment(qualityScale);
  if (adjust) {
    saveInfo.quality 
      = (int) ( (float) adjust->value / 100.0 * 256.0 );
  }

  /* parse the base filename and extension */
  strcpy( filename, file);

  /* make sure a file extension is supplied */
  char *tmp = strrchr( filename, '.');
  if (tmp == NULL) {
    modal_message( "You must enter a filename with an extension" );
    return;
  } else {
    tmp[0] = '\0';
    strcpy( extension, strrchr( file, '.'));
  }
    
  /* Iterate over all frames in selection */
  for (int i = begin; i <= end && exportPage->isExporting; i += every ) {

    /* Call innerLoopUpdate */
    innerLoopUpdate(i, begin, end, every);
      
    /* Extract pixel data, save to image */
    playlist->GetFrame( i, frame );
    frame.ExtractRGB( pixels );
    image = gdk_imlib_create_image_from_data
      (pixels, NULL, frame.GetWidth(), frame.GetHeight());
    sprintf( file, "%s_%08i%s", filename, i, extension);
    gdk_imlib_save_image( image, file, &saveInfo );
    gdk_imlib_destroy_image( image );
  }
} 
