/*
 * page_magick.cc -- Magick Page definition
 * Copyright (C) 2002 Charles Yates <charles.yates@pandora.be>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <vector>
using std::cout;
using std::endl;
using std::vector;

#include <string>
#include <strstream>
#include <iostream>
#include <iomanip>
using std::strstream;
using std::setw;
using std::setfill;

#include "kino_common.h"
#include "page.h"

#include "page_magick.h"
#include "page_editor.h"
#include "displayer.h"
#include "message.h"
#include "error.h"

extern KinoCommon *common;

extern "C" {
#include "magick_interface.h"
#include "support.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>

static bool buttonMutex = false;

void
on_togglebutton_magick_preview_toggled (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	if ( !buttonMutex ) 
	{
		buttonMutex = true;
		GtkNotebook *notebook = GTK_NOTEBOOK( lookup_widget( common->getPageMagick()->window, "notebook_magick" ) );
		gtk_notebook_set_page( notebook, 3 );
		common->getPageMagick()->StartPreview();
		buttonMutex = false;
	}
}

void
on_togglebutton_magick_start_toggled   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	if ( !buttonMutex ) 
	{
		buttonMutex = true;
		common->getPageMagick()->StartRender();
		buttonMutex = false;
	}
}

void
on_togglebutton_magick_stop_toggled    (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	if ( !buttonMutex ) 
	{
		buttonMutex = true;
		common->getPageMagick()->Stop();
		buttonMutex = false;
	}
}

gboolean
on_eventbox_magick_event               (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	common->getPageMagick()->RefreshStatus();
	return FALSE;
}

void
on_notebook_magick_switch_page         (GtkNotebook     *notebook,
                                        GtkNotebookPage *page,
                                        gint             page_num,
                                        gpointer         user_data)
{
	if ( page_num == 3 )
		common->getPageMagick()->LoadSplash();
	else
		common->getPageMagick()->StopPreview();
}

gboolean
on_hscale_transition_start_button_press_event
                                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
	GtkWidget *lower = lookup_widget (common->getPageMagick()->window, "hscale_transition_start");
	GtkWidget *upper = lookup_widget (common->getPageMagick()->window, "hscale_transition_end");
	GtkAdjustment *adjust_lower = GTK_RANGE( lower )->adjustment;
	GtkAdjustment *adjust_upper = GTK_RANGE( upper )->adjustment;
	adjust_upper->lower = adjust_lower->value;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust_upper ), "changed" );
  	return FALSE;
}


gboolean
on_hscale_transition_start_button_release_event
                                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
	GtkWidget *lower = lookup_widget (common->getPageMagick()->window, "hscale_transition_start");
	GtkWidget *upper = lookup_widget (common->getPageMagick()->window, "hscale_transition_end");
	GtkAdjustment *adjust_lower = GTK_RANGE( lower )->adjustment;
	GtkAdjustment *adjust_upper = GTK_RANGE( upper )->adjustment;
	adjust_upper->lower = adjust_lower->value;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust_upper ), "changed" );
  	return FALSE;
}


gboolean
on_hscale_transition_end_button_press_event
                                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
	GtkWidget *lower = lookup_widget (common->getPageMagick()->window, "hscale_transition_start");
	GtkWidget *upper = lookup_widget (common->getPageMagick()->window, "hscale_transition_end");
	GtkAdjustment *adjust_lower = GTK_RANGE( lower )->adjustment;
	GtkAdjustment *adjust_upper = GTK_RANGE( upper )->adjustment;
	adjust_lower->upper = adjust_upper->value;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust_lower ), "changed" );
  	return FALSE;
}


gboolean
on_hscale_transition_end_button_release_event
                                        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{

	GtkWidget *lower = lookup_widget (common->getPageMagick()->window, "hscale_transition_start");
	GtkWidget *upper = lookup_widget (common->getPageMagick()->window, "hscale_transition_end");
	GtkAdjustment *adjust_lower = GTK_RANGE( lower )->adjustment;
	GtkAdjustment *adjust_upper = GTK_RANGE( upper )->adjustment;
	adjust_lower->upper = adjust_upper->value;
	gtk_signal_emit_by_name( GTK_OBJECT( adjust_lower ), "changed" );
  	return FALSE;
}

void
on_togglebutton_advanced_toggled       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
	GtkWidget *frame = lookup_widget( common->getPageMagick()->window, "frame_advanced" );
	if ( gtk_toggle_button_get_active( togglebutton ) )
		gtk_widget_show_now( frame );
	else
		gtk_widget_hide( frame );
}

}

/** Plugin - wrapper for the dlopen/dlsym functions.
*/

bool Plugin::Open( char *file )
{
	ptr = dlopen( file, RTLD_NOW );
	return ptr != NULL;
}

void Plugin::Close( )
{
	if ( ptr != NULL )
		dlclose( ptr );
}

void *Plugin::Find( char *symbol )
{
	return dlsym( ptr, symbol );
}

char *Plugin::GetError( )
{
	return dlerror();
}

/** Plugin Collection - loads all shared objects in the specified directory.
*/

PluginCollection::PluginCollection( )
{
}

PluginCollection::~PluginCollection( )
{
	for ( unsigned int index = 0; index < collection.size(); index ++ )
	{
		collection[ index ]->Close();
		delete collection[ index ];
	}
}

void PluginCollection::Initialise( char *directory )
{
	char *filename; 
	char *extension;
	DIR *dir;
	struct dirent *entry;
	struct stat statbuf;

	dir = opendir( directory );

	if ( dir )
	{
		while ( ( entry = readdir( dir ) ) != NULL )
		{
			filename = g_strdup_printf( "%s/%s", directory, entry->d_name );
			extension = strrchr( entry->d_name, '.' );
			if ( extension != NULL && !stat( filename, &statbuf ) && S_ISREG( statbuf.st_mode ) )
			{
				if ( !strcmp( extension, ".so" ) )
				{
					RegisterPlugin( filename );
				}
			}
			g_free(filename);
		}
		closedir(dir);
	}
}

void PluginCollection::RegisterPlugin( char *filename )
{
	Plugin *plugin = new Plugin;
	if ( plugin->Open( filename ) )
	{
		cout << ">>> Registering plugin " << filename << endl;
		collection.push_back( plugin );
	}
	else
	{
		cout << ">>> Rejecting plugin " << filename << " : " << plugin->GetError() << endl;
		delete plugin;
	}
}

unsigned int PluginCollection::Count()
{
	return collection.size();
}

Plugin *PluginCollection::Get( unsigned int index )
{
	return collection[ index ];
}

void PluginImageCreateRepository::InstallPlugins( Plugin *plugin )
{
	GDKImageCreate *( *func )( int ) = ( GDKImageCreate *( * )( int ) )plugin->Find( "GetImageCreate" );
	if ( func != NULL )
	{
		int index = 0;
		GDKImageCreate *entry = func( index ++ );
		while ( entry != NULL )
		{
			Register( entry );
			entry = func( index ++ );
		}
	}
}

void PluginImageFilterRepository::InstallPlugins( Plugin *plugin )
{
	GDKImageFilter *( *func )( int ) = ( GDKImageFilter *( * )( int ) )plugin->Find( "GetImageFilter" );
	if ( func != NULL )
	{
		int index = 0;
		GDKImageFilter *entry = func( index ++ );
		while ( entry != NULL )
		{
			Register( entry );
			entry = func( index ++ );
		}
	}
}

void PluginImageTransitionRepository::InstallPlugins( Plugin *plugin )
{
	GDKImageTransition *( *func )( int ) = ( GDKImageTransition *( * )( int ) )plugin->Find( "GetImageTransition" );
	if ( func != NULL )
	{
		int index = 0;
		GDKImageTransition *entry = func( index ++ );
		while ( entry != NULL )
		{
			Register( entry );
			entry = func( index ++ );
		}
	}
}

void PluginAudioFilterRepository::InstallPlugins( Plugin *plugin )
{
	GDKAudioFilter *( *func )( int ) = ( GDKAudioFilter *( * )( int ) )plugin->Find( "GetAudioFilter" );
	if ( func != NULL )
	{
		int index = 0;
		GDKAudioFilter *entry = func( index ++ );
		while ( entry != NULL )
		{
			Register( entry );
			entry = func( index ++ );
		}
	}
}

void PluginAudioTransitionRepository::InstallPlugins( Plugin *plugin )
{
	GDKAudioTransition *( *func )( int ) = ( GDKAudioTransition *( * )( int ) )plugin->Find( "GetAudioTransition" );
	if ( func != NULL )
	{
		int index = 0;
		GDKAudioTransition *entry = func( index ++ );
		while ( entry != NULL )
		{
			Register( entry );
			entry = func( index ++ );
		}
	}
}

/** Common info and factory for all PageMagick processors.
*/

class PageMagickFrames;
class PageMagickImage;
class PageMagickAudio;

class PageMagickInfo 
{
	private:
		PageMagickFrames *frameSource;
		PageMagickImage *imageManipulator;
		PageMagickAudio *audioManipulator;
	public:
		KinoCommon *common;
		int begin;
		int end;
		double increment;
		bool reverse;
		int anteFrame;
		int postFrame;
		int width;
		int height;
		int isPAL;
		int isWide;
		int frequency;
		short channels;
		int samples_this_frame;
		PageMagickInfo( KinoCommon *common );
		~PageMagickInfo();
		KinoCommon *GetCommon( ) { return this->common; }
		void Initialise();
		PageMagickFrames *GetFrameSource();
		PageMagickImage *GetImageManipulator();
		PageMagickAudio *GetAudioManipulator();
		void SetBegin( int begin ) { this->begin = begin > 0 ? begin : 0; }
		void SetEnd( int end ) { this->end = end > 0 ? end : 0; }
		void SetPostFrame( int postFrame ) { this->postFrame = postFrame; }
		void SetAnteFrame( int anteFrame ) { this->anteFrame = anteFrame; }
		void GetAnteFrame( uint8_t *pixels );
		void GetPostFrame( uint8_t *pixels );
		int GetPostFrame( ) { return this->postFrame; }
		int GetAnteFrame( ) { return this->anteFrame; }
};

/** PageMagick frame source - this interface is used to source the frames.
*/

class PageMagickFrames 
{
	public:
		virtual void Initialise( PageMagickInfo * ) { }
		virtual void GetFrame( uint8_t *pixels, int16_t **audio, int i ) { }
		virtual void Close() { }
};

/** PageMagic Image manipulator - this interface is used to manipulate the images.
*/

class PageMagickImage 
{
	public:
		virtual void Initialise( PageMagickInfo *info ) { };
		virtual void PreGetFrame( ) { };
		virtual void GetFrame( uint8_t *pixels, int i ) { };
		virtual void Close() { };
};

/** PageMagic Audio manipulator - this interface is used to manipulate the audio.
*/

class PageMagickAudio 
{
	public:
		virtual void Initialise( PageMagickInfo * ) { };
		virtual void GetFrame( int16_t **audio, int i ) { };
		virtual void Close() { };
};

static Frame *infoFrame = new Frame;

/** Implementation of the Overwrite frame source.
*/

class PageMagickOverwrite : public PageMagickFrames 
{
	private:
		PageMagickInfo *info;
	public:
		void Initialise( PageMagickInfo * );
		void GetFrame( uint8_t *pixels, int16_t **audio, int i );
		void Close();
};

void PageMagickOverwrite::Initialise( PageMagickInfo *info ) 
{
	this->info = info;
	// Obtain begining and ending of sequence
	info->SetBegin( 0 );
	info->SetEnd( info->GetCommon()->getPlayList()->GetNumFrames() );
	info->increment = 1;
	info->reverse = false;

	GtkToggleButton *rangeToggle = GTK_TOGGLE_BUTTON( lookup_widget( info->GetCommon()->getPageMagick()->window, "radiobutton_magick_range" ) );
	GtkEntry *startSpin = GTK_ENTRY( lookup_widget( info->GetCommon()->getPageMagick()->window, "spinbutton_magick_start" ) );
	GtkEntry *endSpin = GTK_ENTRY( lookup_widget( info->GetCommon()->getPageMagick()->window, "spinbutton_magick_end" ) );	

	if ( gtk_toggle_button_get_active( rangeToggle ) ) 
	{
		info->SetBegin( atoi( gtk_entry_get_text( startSpin ) ) - 1 );
		info->SetEnd( atoi( gtk_entry_get_text( endSpin ) ) - 1 );
		GtkToggleButton *limit = GTK_TOGGLE_BUTTON( lookup_widget( info->GetCommon()->getPageMagick()->window, "checkbutton_magick_frame_limit" ) );
		if ( gtk_toggle_button_get_active( limit ) ) 
		{
			GtkEntry *spin = GTK_ENTRY( lookup_widget( info->GetCommon()->getPageMagick()->window, "spinbutton_magick_frame_count" ) );	
        	GtkMenu *menu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( info->GetCommon()->getPageMagick()->window, "optionmenu_magick_frame_offset" ) ) ) );
        	GtkWidget *active_item = gtk_menu_get_active( menu );
        	if ( g_list_index (GTK_MENU_SHELL ( menu )->children, active_item) == 0 ) 
			{
				int end = info->begin + atoi( gtk_entry_get_text( spin ) );
				if ( end < info->end )
					info->SetEnd( end );
			}
			else 
			{
				info->SetBegin( info->end - atoi( gtk_entry_get_text( spin ) ) );
			}
		}
		info->SetAnteFrame( info->begin - 1 );
		info->SetPostFrame( info->end + 1 );
	}
	else 
	{
		throw "No frames selected for rewriting.";
	}

	GtkToggleButton *advanced = GTK_TOGGLE_BUTTON( lookup_widget( info->GetCommon()->getPageMagick()->window, "togglebutton_advanced" ) );

	if ( gtk_toggle_button_get_active( advanced ) )
	{
		// Determine speed
		GtkToggleButton *speed = GTK_TOGGLE_BUTTON( lookup_widget( info->GetCommon()->getPageMagick()->window, "checkbutton_speed" ) );
		if ( gtk_toggle_button_get_active( speed ) )
		{
			GtkRange *range = GTK_RANGE( lookup_widget( info->GetCommon()->getPageMagick()->window, "hscale_speed" ) );
			info->increment = range->adjustment->value;
		}

		// Determine direction
		GtkToggleButton *reverse = GTK_TOGGLE_BUTTON( lookup_widget( info->GetCommon()->getPageMagick()->window, "checkbutton_reverse" ) );
		info->reverse = gtk_toggle_button_get_active( reverse );
	}
}

void PageMagickOverwrite::GetFrame( uint8_t *pixels, int16_t **audio, int i ) 
{
	common->getPlayList()->GetFrame( i, *( infoFrame ) );
	infoFrame->ExtractRGB( pixels );
	infoFrame->ExtractAudio( audio );
	info->samples_this_frame = infoFrame->decoder->audio->samples_this_frame;
}

void PageMagickOverwrite::Close( ) 
{
	cout << "Deleting " << info->begin << " << " << info->end << endl;
	info->GetCommon()->getPlayList()->Delete( info->begin, info->end );
}

/** Implementation of the Create frame source.
*/

class PageMagickCreate : public PageMagickFrames 
{
	private:
		PageMagickInfo *info;
		uint8_t *image;
		GDKImageCreate *creator;

	public:
		PageMagickCreate();
		virtual ~PageMagickCreate();
		void Initialise( PageMagickInfo * );
		void GetFrame( uint8_t *pixels, int16_t **audio, int i );
};

PageMagickCreate::PageMagickCreate() : image( NULL ) 
{
	image = new uint8_t[ 720 * 576 * 3 ];
}

PageMagickCreate::~PageMagickCreate() 
{
	delete image;
}

void PageMagickCreate::Initialise( PageMagickInfo *info ) 
{
	this->info = info;
	info->begin = common->g_currentFrame;
	info->increment = 1;
	info->reverse = false;

	GtkToggleButton *frameToggle = GTK_TOGGLE_BUTTON( lookup_widget( common->getPageMagick()->window, "radiobutton_magick_create_before" ) );
	if ( gtk_toggle_button_get_active( frameToggle ) ) 
	{
		GtkEntry *spin = GTK_ENTRY( lookup_widget( common->getPageMagick()->window, "spinbutton_magick_create_before" ) );
		info->begin = atoi( gtk_entry_get_text( spin ) ) - 1;
	}

	info->anteFrame = info->begin - 1;
	info->postFrame = info->begin + 1;

	creator = info->GetCommon()->getPageMagick()->GetImageCreate();
	if ( creator != NULL )
		creator->InterpretWidgets( GTK_BIN( lookup_widget( common->getPageMagick()->window, "frame_magick_frames_create" ) ) );
	else
		throw "Invalid image creator selected";

	info->end = info->begin + creator->GetNumberOfFrames() - 1;
}

void PageMagickCreate::GetFrame( uint8_t *pixels, int16_t **audio, int i ) 
{
	double position = info->end == info->begin ? 0 : (double)( i - info->begin ) / (double)( info->end - info->begin );
	const double frame_delta = ( info->end == info->begin ? 0.0 : 1.0 / static_cast<double>(info->end - info->begin) ) * info->increment;
	
	if ( creator != NULL )
		creator->CreateFrame( image, info->width, info->height, position, frame_delta );
	else
		throw "Invalid image creator selected";

	memcpy( pixels, image, info->width * info->height * 3 );

	for ( int i = 0; i < 4; i ++ )
		memset( audio[ i ], 0, 2 * DV_AUDIO_MAX_SAMPLES * sizeof(int16_t) );
}

/** Filter implementation
*/

class PageMagickTransition : public PageMagickImage 
{
	private:
		PageMagickInfo *info;
		GDKImageTransition *transition;
		uint8_t keyFrame[ 720 * 576 * 3 ];
		bool animateKey;
		int direction;
		double start_position;
		double end_position;
	public:
		void Initialise( PageMagickInfo *info );
		void PreGetFrame( );
		void GetFrame( uint8_t *pixels, int i );
		void Close();
};

void PageMagickTransition::Initialise( PageMagickInfo *info ) 
{
	this->info = info;
	this->animateKey = false;

	transition = info->GetCommon()->getPageMagick()->GetImageTransition();
	if ( transition != NULL )
		transition->InterpretWidgets( GTK_BIN( lookup_widget( common->getPageMagick()->window, "frame_magick_image_transition" ) ) );
	else
		throw "Invalid image transition selected";

 	GtkToggleButton *toggle = GTK_TOGGLE_BUTTON( lookup_widget( common->getPageMagick()->window, "radiobutton_magick_transition_colour" ) );
	
	if ( gtk_toggle_button_get_active( toggle ) ) 
	{
		GnomeColorPicker *colourPicker = GNOME_COLOR_PICKER( lookup_widget( info->GetCommon()->getPageMagick()->window, "colorpicker_magick_transition" ) );
    	guint8 r, g, b, a;
    	gnome_color_picker_get_i8( colourPicker, &r, &g, &b, &a);
		for ( uint8_t *p = keyFrame; p < ( keyFrame + info->width * info->height * 3 ); ) 
		{
			*p ++ = r;
			*p ++ = g;
			*p ++ = b;
		}
	}
	else 
	{
		GtkMenu *menu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( info->GetCommon()->getPageMagick()->window, "optionmenu_magick_transition_frame" ) ) ) );
		GtkWidget *active_item = gtk_menu_get_active( menu );
		switch( g_list_index (GTK_MENU_SHELL ( menu )->children, active_item) ) {
			case 0:
				info->GetPostFrame( keyFrame );
				break;
			case 1:
				info->GetAnteFrame( keyFrame );
				break;
			case 2:
				animateKey = true;
				break;
		}
	}

	GtkMenu *menu = GTK_MENU( gtk_option_menu_get_menu( GTK_OPTION_MENU( lookup_widget( info->GetCommon()->getPageMagick()->window, "optionmenu_direction" ) ) ) );
	GtkWidget *active_item = gtk_menu_get_active( menu );
	direction = g_list_index (GTK_MENU_SHELL ( menu )->children, active_item);

	GtkRange *range = GTK_RANGE( lookup_widget( info->GetCommon()->getPageMagick()->window, "hscale_transition_start" ) );
	start_position = range->adjustment->value;
	range = GTK_RANGE( lookup_widget( info->GetCommon()->getPageMagick()->window, "hscale_transition_end" ) );
	end_position = range->adjustment->value;
}

void PageMagickTransition::PreGetFrame( ) 
{
	if ( animateKey ) 
	{
		info->GetPostFrame( keyFrame );
		info->postFrame ++;
	}
}

void PageMagickTransition::GetFrame( uint8_t *pixels, int i ) 
{
	double position = info->end == info->begin ? 1 : (double)( i - info->begin ) / (double)( info->end - info->begin );

	position = start_position + ( end_position - start_position ) * position;

	const double frame_delta = ( info->end == info->begin ? 1.0 : 1.0 / static_cast<double>(info->end - info->begin) ) * info->increment;

	if ( transition != NULL )
		transition->GetFrame( pixels, keyFrame, info->width, info->height, position, frame_delta, direction == 1 );
	else
		throw "Invalid image transition selected";
}

void PageMagickTransition::Close( ) 
{
	if ( this->animateKey )
	{
		cout << "Deleting " << info->begin << " << " << info->begin + ( info->end - info->begin ) / info->increment << endl;
		info->GetCommon()->getPlayList()->Delete( info->begin, (int)( info->begin + ( info->end - info->begin ) / info->increment ) );
	}
}

/** Filter implementation
*/

class PageMagickFilter : public PageMagickImage 
{
	private:
		PageMagickInfo *info;
		GDKImageFilter *filter;
	public:
		void Initialise( PageMagickInfo *info );
		void GetFrame( uint8_t *pixels, int i );
};

void PageMagickFilter::Initialise( PageMagickInfo *info ) 
{
	this->info = info;
	filter = info->GetCommon()->getPageMagick()->GetImageFilter();
	if ( filter != NULL )
		filter->InterpretWidgets( GTK_BIN( lookup_widget( common->getPageMagick()->window, "frame_magick_image_filter" ) ) );
	else
		throw "Invalid image filter selected";
}

void PageMagickFilter::GetFrame( uint8_t *pixels, int i ) 
{
	double position = info->end == info->begin ? 0.5 : (double)( i - info->begin ) / (double)( info->end - info->begin );
	const double frame_delta = ( info->end == info->begin ? 0.0 : 1.0 / static_cast<double>(info->end - info->begin ) ) * info->increment;
	
	if ( filter != NULL )
		filter->FilterFrame( pixels, info->width, info->height, position, frame_delta );
	else
		throw "Invalid image filter selected";
}

/** PageMagic Audio Filter implementation. 
*/

class PageMagickAudioFilter : public PageMagickAudio 
{
	private:
		PageMagickInfo *info;
		GDKAudioFilter *filter;
	public:
		void Initialise( PageMagickInfo *info );
		void GetFrame( int16_t **audio, int i );
};

void PageMagickAudioFilter::Initialise( PageMagickInfo *info ) 
{
	this->info = info;
	filter = info->GetCommon()->getPageMagick()->GetAudioFilter();
	if ( filter != NULL )
		filter->InterpretWidgets( GTK_BIN( lookup_widget( common->getPageMagick()->window, "frame_magick_audio_filter" ) ) );
	else
		throw "Invalid audio filter selected";
}

void PageMagickAudioFilter::GetFrame( int16_t **audio, int i ) 
{
	double position = info->end == info->begin ? 0.5 : (double)( i - info->begin ) / (double)( info->end - info->begin );
	const double frame_delta = ( info->end == info->begin ? 0.0 : 1.0 / static_cast<double>(info->end - info->begin) ) * info->increment;

	if ( filter != NULL )
		filter->GetFrame( audio, info->frequency, info->channels, info->frequency / ( info->isPAL ? 25 : 30 ), position, frame_delta );
	else
		throw "Invalid audio filter selected";
}

/** PageMagick Audio Transition implementation. 
*/

class PageMagickAudioTransition : public PageMagickAudio 
{
	private:
		PageMagickInfo *info;
		int16_t *audio_buffers[ 4 ];
		GDKAudioTransition *transition;

	public:
		PageMagickAudioTransition();
		virtual ~PageMagickAudioTransition();
		void Initialise( PageMagickInfo *info );
		void GetFrame( int16_t **audio, int i );
};

PageMagickAudioTransition::PageMagickAudioTransition()
{
	for ( int index = 0; index < 4; index ++ )
		audio_buffers[ index ] = new int16_t[ 2 * DV_AUDIO_MAX_SAMPLES ];
}

PageMagickAudioTransition::~PageMagickAudioTransition()
{
	for ( int index = 0; index < 4; index ++ )
		delete audio_buffers[ index ];
}

void PageMagickAudioTransition::Initialise( PageMagickInfo *info ) 
{
	this->info = info;

	transition = info->GetCommon()->getPageMagick()->GetAudioTransition();
	if ( transition != NULL )
		transition->InterpretWidgets( GTK_BIN( lookup_widget( common->getPageMagick()->window, "frame_magick_audio_transition" ) ) );
	else
		throw "Invalid audio transition selected";
}

void PageMagickAudioTransition::GetFrame( int16_t **audio, int i ) 
{
	// Pick up the b-frame audio (ugly.. moved by image transition when appropriate)
	// should also be resampled to the a-frames frequency - this is going to go haywire in
	// mixed sample rate projects :-/
	// In truth, even the aframes should be resampled to original aframes frequency... (fun, fun, fun)
	infoFrame->ExtractAudio( audio_buffers );

	double position = info->end == info->begin ? 0 : (double)( i - info->begin ) / (double)( info->end - info->begin );
	const double frame_delta = ( info->end == info->begin ? 0.0 : 1.0 / static_cast<double>(info->end - info->begin) ) * info->increment;
	
	if ( transition != NULL )
		transition->GetFrame( audio, audio_buffers, info->frequency, info->channels, info->frequency / ( info->isPAL ? 25 : 30 ), position, frame_delta );
	else
		throw "Invalid audio filter selected";
}

/** Info implementation.
*/

PageMagickInfo::PageMagickInfo( KinoCommon *common ) 
{
	this->common = common;
	this->frameSource = NULL;
	this->imageManipulator = NULL;
	this->audioManipulator = NULL;
}

PageMagickInfo::~PageMagickInfo() 
{
	delete frameSource;
	delete imageManipulator;
	delete audioManipulator;
}

void PageMagickInfo::Initialise() 
{
	// Get a sample frame to obtain recording info
	if ( common->getPlayList()->GetFrame( 0, *infoFrame ) ) 
	{
		// Get all video and audio info required
		width = infoFrame->GetWidth();
		height = infoFrame->GetHeight();
		isPAL = infoFrame->IsPAL();
		isWide = infoFrame->IsWide();
		channels = (short)infoFrame->decoder->audio->num_channels;
		frequency = infoFrame->decoder->audio->frequency;
	}
	else if ( Preferences::getInstance().defaultNormalisation != NORM_UNSPECIFIED ) 
	{
		// Get default video and audio info from preferences.
		width = 720;
		height = Preferences::getInstance().defaultNormalisation == NORM_PAL ? 576 : 480;
		isPAL = Preferences::getInstance().defaultNormalisation == NORM_PAL;
		isWide = Preferences::getInstance().defaultAspect == ASPECT_169;
		channels = (short)2;
		switch( Preferences::getInstance().defaultAudio ) 
		{
			case AUDIO_32KHZ: frequency = 32000; break;
			case AUDIO_44KHZ: frequency = 44100; break;
			case AUDIO_48KHZ: frequency = 48000; break;
		}
	}
	else 
	{
		throw "The playlist is empty and the default preferences for video creation have not specified - aborting.";
	}
}

PageMagickFrames *PageMagickInfo::GetFrameSource() 
{
	if ( frameSource == NULL ) 
	{
		GtkNotebook *notebook = GTK_NOTEBOOK( lookup_widget( this->common->getPageMagick()->window, "notebook_magick_frames" ) );
		int page = gtk_notebook_get_current_page( notebook );

		switch( page ) 
		{
			case 0:
				frameSource = new PageMagickOverwrite();
				break;
			case 1:
				frameSource = new PageMagickCreate();
				break;
		}
	}
	
	if ( frameSource != NULL ) 
		frameSource->Initialise( this );
	else 
		throw "The requested frame source has not been implemented yet...";

	return frameSource;
}

PageMagickImage *PageMagickInfo::GetImageManipulator() 
{
	if ( imageManipulator == NULL ) 
	{
		GtkNotebook *notebook = GTK_NOTEBOOK( lookup_widget( this->common->getPageMagick()->window, "notebook_magick_video" ) );
		int page = gtk_notebook_get_current_page( notebook );

		switch( page ) 
		{
			case 0:
				imageManipulator = new PageMagickFilter();
				break;
			case 1:
				imageManipulator = new PageMagickTransition();
				break;
		}
	}

	if ( imageManipulator != NULL ) 
		imageManipulator->Initialise( this );
	else 
		throw "The requested image manipulator has not been implemented yet...";

	return imageManipulator;
}

PageMagickAudio *PageMagickInfo::GetAudioManipulator() 
{
	if ( audioManipulator == NULL ) 
	{
		GtkNotebook *notebook = GTK_NOTEBOOK( lookup_widget( this->common->getPageMagick()->window, "notebook_magick_audio" ) );
		int page = gtk_notebook_get_current_page( notebook );

		switch( page ) 
		{
			case 0:
				audioManipulator = new PageMagickAudioFilter();
				break;
			case 1:
				audioManipulator = new PageMagickAudioTransition();
				break;
		}
	}

	if ( audioManipulator != NULL ) 
		audioManipulator->Initialise( this );
	else 
		throw "The requested audio manipulator has not been implemented yet...";

	return audioManipulator;
}

void PageMagickInfo::GetAnteFrame( uint8_t *pixels ) 
{
	if ( this->anteFrame >= 0 ) 
	{
		common->getPlayList()->GetFrame( this->anteFrame, *infoFrame );
		infoFrame->ExtractRGB( pixels );
	}
	else 
	{
		memset( pixels, 0, 720 * 576 * 3 );
	}
}

void PageMagickInfo::GetPostFrame( uint8_t *pixels ) 
{
	if ( this->postFrame < common->getPlayList()->GetNumFrames() ) 
	{
		common->getPlayList()->GetFrame( this->postFrame, *infoFrame );
		infoFrame->ExtractRGB( pixels );
	}
	else 
	{
		memset( pixels, 0, 720 * 576 * 3 );
	}
}

/** Constructor for Page Magick.
*/

PageMagick::PageMagick( KinoCommon *common ) : rendering( false ), previewing( false ) 
{ 
	cout << "> Creating Magick Page" << endl;

	window = create_window_magick( );
	GtkBin *bin = GTK_BIN( lookup_widget( common->getWidget(), "frame_magick" ) );
	gtk_widget_reparent( ( GTK_BIN( window ) )->child, GTK_WIDGET( bin ) );

	this->common = common;

	progressBar = GTK_PROGRESS_BAR( lookup_widget( window, "progressbar_magick" ) );

	plugins.Initialise( KINO_PLUGINDIR );

	for ( unsigned int index = 0; index < plugins.Count(); index ++ )
	{
		image_creators.InstallPlugins( plugins.Get( index ) );
		image_filters.InstallPlugins( plugins.Get( index ) );
		image_transitions.InstallPlugins( plugins.Get( index ) );
		audio_filters.InstallPlugins( plugins.Get( index ) );
		audio_transitions.InstallPlugins( plugins.Get( index ) );
	}

	GtkOptionMenu *menu = GTK_OPTION_MENU( lookup_widget( window, "optionmenu_magick_filter" ) );
	GtkBin *container = GTK_BIN( lookup_widget( window, "frame_magick_image_filter" ) );
	image_filters.Initialise( menu, container );

	menu = GTK_OPTION_MENU( lookup_widget( window, "optionmenu_magick_frames_create" ) );
	container = GTK_BIN( lookup_widget( window, "frame_magick_frames_create" ) );
	image_creators.Initialise( menu, container );

	menu = GTK_OPTION_MENU( lookup_widget( window, "optionmenu_magick_transition" ) );
	container = GTK_BIN( lookup_widget( window, "frame_magick_image_transition" ) );
	image_transitions.Initialise( menu, container );

	menu = GTK_OPTION_MENU( lookup_widget( window, "optionmenu_magick_audio_filter" ) );
	container = GTK_BIN( lookup_widget( window, "frame_magick_audio_filter" ) );
	audio_filters.Initialise( menu, container );

	menu = GTK_OPTION_MENU( lookup_widget( window, "optionmenu_magick_audio_transition" ) );
	container = GTK_BIN( lookup_widget( window, "frame_magick_audio_transition" ) );
	audio_transitions.Initialise( menu, container );
}

/** Destructor for Page Magick.
*/

PageMagick::~PageMagick() 
{
	GtkBin *bin = GTK_BIN( lookup_widget( common->getWidget(), "frame_magick" ) );
	gtk_widget_reparent( ( GTK_BIN( bin ) )->child, GTK_WIDGET( window ) );
	gtk_widget_destroy( window );
}

/** Start when entering page.
*/

void PageMagick::start() 
{
	cout << ">>> Starting magick" << endl;
	strcpy( status, "" );
	RefreshStatus();
}

/** Activate scene list selection (for Frame sources).
*/

gulong PageMagick::activate() 
{
	return SCENE_LIST;
}

/** Clean up when leaving page.
*/

void PageMagick::clean() 
{
	StopPreview();
	strcpy( status, "" );
}

/** Play an audio frame.
*/

void PageMagick::PlayAudio( int16_t *buffers[], int frequency, int channels, bool isPAL ) 
{
	dv_audio.frequency = frequency;
	dv_audio.samples_this_frame = frequency / ( isPAL ? 25 : 30 );
	dv_audio.num_channels = channels;

	if ( !audio_device_avail && ( audio_sampling_rate = dv_oss_init( &dv_audio, audio_device ) ) != 0 )
		audio_device_avail = true;
	if ( audio_device_avail )
		dv_oss_play( &dv_audio, audio_device, buffers );
}

/** Show an image in the preview drawing area.
*/

void PageMagick::ShowImage( GtkWidget *area, uint8_t *pixels, int width, int height ) 
{
	GdkImlibImage *img = gdk_imlib_create_image_from_data( pixels, NULL, width, height );
	gdk_imlib_apply_image( img, area->window );
	gdk_imlib_destroy_image( img );
}

/** Provide a preview of the effect.
*/

void PageMagick::StartPreview() 
{
	// Make sure we're not already rendering or previewing.
	if ( rendering || previewing )
		return;

	buttonMutex = true;
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_preview" ) ), 1 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_start" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_stop" ) ), 0 );
	buttonMutex = false;

	// Set the page condition
	previewing = true;

	GtkWidget *area = GTK_WIDGET( lookup_widget( window, "drawingarea_magick_preview" ) );

	// Create the temporary space
	uint8_t *pixels = new uint8_t[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];
	int16_t *audio_buffers[ 4 ];
	for ( int n = 0; n < 4; n++ )
		audio_buffers[ n ] = new int16_t [2 * DV_AUDIO_MAX_SAMPLES];

	// Generate the info
	PageMagickInfo *info = new PageMagickInfo( common );

	GtkToggleButton *audioButton = GTK_TOGGLE_BUTTON( lookup_widget( window, "checkbutton_magick_preview_audio" ) );
	GtkToggleButton *everyButton = GTK_TOGGLE_BUTTON( lookup_widget( window, "checkbutton_magick_preview_every" ) );
	GtkEntry *everySpin = GTK_ENTRY( lookup_widget( window, "spinbutton_magick_preview_every" ) );
	GtkToggleButton *loopButton = GTK_TOGGLE_BUTTON( lookup_widget( window, "checkbutton_magick_preview_loop" ) );
	GtkToggleButton *contextButton = GTK_TOGGLE_BUTTON( lookup_widget( window, "checkbutton_magick_preview_context" ) );
	GtkEntry *contextSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_magick_preview_context" ) );

	audio_device = dv_oss_new();
	audio_device_avail = false;
	audio_sampling_rate = 0;

	try 
	{
		do 
		{
			info->Initialise();

			PageMagickFrames *frames = info->GetFrameSource();
			PageMagickImage *image = info->GetImageManipulator();
			PageMagickAudio *sound = info->GetAudioManipulator();

			if ( info->begin > info->end )
				throw "Invalid frame range specified.";

			for ( int i = info->begin - atoi( gtk_entry_get_text( contextSpin ) );
				  i < info->begin && gtk_toggle_button_get_active( contextButton ) && previewing;
				  i ++ ) 
			{
				if ( i >= 0 ) {
					common->getPlayList()->GetFrame( i, *( infoFrame ) );

					if ( !gtk_toggle_button_get_active( everyButton ) || i % atoi( gtk_entry_get_text( everySpin ) ) == 0 ) 
					{
						infoFrame->ExtractRGB( pixels );
						ShowImage( area, pixels, info->width, info->height );
					}

					if ( gtk_toggle_button_get_active( audioButton ) ) 
					{
						infoFrame->ExtractAudio( audio_buffers );
						PlayAudio( audio_buffers, info->frequency, info->channels, info->isPAL );
					}

					UpdateProgress( (gfloat) 0 );
				}
			}

			for (double i = (double)info->begin; i < (double)( info->end + 1 ) && previewing; i += info->increment ) 
			{
				int frame_number = (int)i;
				if ( info->reverse )
					frame_number = info->end - ( (int)i - info->begin );

				frames->GetFrame( pixels, audio_buffers, frame_number );
				image->PreGetFrame( );

				if ( !gtk_toggle_button_get_active( everyButton ) || 
					( i == info->begin || ( frame_number - info->begin ) % atoi( gtk_entry_get_text( everySpin ) ) == 0 ) ) 
				{
					image->GetFrame( pixels, frame_number );
					ShowImage( area, pixels, info->width, info->height );
				}

				if ( gtk_toggle_button_get_active( audioButton ) ) 
				{
					sound->GetFrame( audio_buffers, frame_number );
					PlayAudio( audio_buffers, info->frequency, info->channels, info->isPAL );
				}

				if (info->end != info->begin)
					UpdateProgress( (gfloat) (( i - info->begin ) ) / (gfloat) (info->end - info->begin) );
			}

			for ( int i = info->postFrame; 
				  i < info->postFrame + atoi( gtk_entry_get_text( contextSpin ) ) && gtk_toggle_button_get_active( contextButton ) && previewing;
				  i ++ ) 
			{
				if ( i < common->getPlayList()->GetNumFrames() ) 
				{
					common->getPlayList()->GetFrame( i, *( infoFrame ) );

					if ( !gtk_toggle_button_get_active( everyButton ) || i % atoi( gtk_entry_get_text( everySpin ) ) == 0 ) 
					{
						infoFrame->ExtractRGB( pixels );
						ShowImage( area, pixels, info->width, info->height );
					}

					if ( gtk_toggle_button_get_active( audioButton ) ) 
					{
						infoFrame->ExtractAudio( audio_buffers );
						PlayAudio( audio_buffers, info->frequency, info->channels, info->isPAL );
					}
				}
				UpdateProgress( (gfloat) 1 );
			}

			UpdateProgress( (gfloat) 1 );
		}
		while( previewing && gtk_toggle_button_get_active( loopButton ) );
	}
	catch( const char *exc ) 
	{
		modal_message( (char *)exc );
	}

	buttonMutex = true;
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_preview" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_start" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_stop" ) ), 1 );
	buttonMutex = false;

	previewing = false;

	dv_oss_close( audio_device );

	delete info;
	delete pixels;
	for ( int n = 0; n < 4; n++ )
		delete audio_buffers[n];
}

/** Render the effect.
*/

void PageMagick::StartRender() 
{
	// Make sure we're not already rendering or previewing already.
	if ( rendering || previewing )
		return;

	// Show the current button status
	buttonMutex = true;
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_preview" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_start" ) ), 1 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_stop" ) ), 0 );
	buttonMutex = false;

	// Set the page condition
	rendering = true;

	// Create the temporary space
	uint8_t *pixels = new unsigned char[FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4];
	int16_t *audio_buffers[ 4 ];
	for ( int n = 0; n < 4; n++ )
		audio_buffers[ n ] = new int16_t [2 * DV_AUDIO_MAX_SAMPLES];
	unsigned char *dv_buffer = new unsigned char[480 * 300];
	unsigned char *dv_pixels[3];
	dv_pixels[ 0 ] = pixels;

	// Build the info
	PageMagickInfo *info = new PageMagickInfo( common );

	// Build the encoder
	dv_encoder_t *encoder = dv_encoder_new( Preferences::getInstance().dvDecoderAddNTSCSetup,
		Preferences::getInstance().dvDecoderClampLuma,
		Preferences::getInstance().dvDecoderClampChroma );

	// Output file
	FILE *output = NULL;

	// All exceptions thrown should be safely picked up in the try/catch
	try 
	{
		// Get the current date and time to set in dv
		time_t datetime = time(NULL);
		int frameNum = 0;
		
		// Initialise the info
		info->Initialise();

		// Obtain the file name
		GtkEntry *fileEntry = GTK_ENTRY( gnome_file_entry_gtk_entry( GNOME_FILE_ENTRY( lookup_widget( window, "fileentry_magick" ) ) ) );
		char *filename = gtk_entry_get_text( fileEntry );
		int counter = 0;
		string thisfile;
		struct stat stats;

		if ( !strcmp( filename, "" ) ) 
			throw "No file name specified";

		// Generate the full file name
		do 
		{
   			strstream       sb;
       		sb << filename << setfill('0') << setw(3) << ++ counter << ".dv";
   			sb >> thisfile;
			cout << ">>> Trying " << thisfile << endl;
		}
		while ( stat( thisfile.c_str(), &stats ) == 0 );

		// This one doesn't exist so use it
		char dvfile[ 1024 ];
		sprintf( dvfile, "%s%03d.dv", filename, counter );

		// Specify the encoder
		encoder->isPAL = (info->height == 576);
		encoder->is16x9 = info->isWide;
		encoder->vlc_encode_passes = 3;
		encoder->static_qno = 0;
		encoder->force_dct = DV_DCT_AUTO;

		// Get the source and manipulators
		PageMagickFrames *frames = info->GetFrameSource();
		PageMagickImage *image = info->GetImageManipulator();
		PageMagickAudio *sound = info->GetAudioManipulator();

		// Check that we're rendering a valid set of frames
		if ( info->begin > info->end )
			throw "Invalid frame range specified.";

		// Open the file
		output = fopen( dvfile, "w" );
		if ( output == NULL )
			throw "Unable to open output file";

		// For each frame
		for (double i = (double)info->begin; i < (double)( info->end + 1 ) && rendering; i += info->increment ) 
		{
			int frame_number = (int)i;
			if ( info->reverse )
				frame_number = info->end - ( (int)i - info->begin );

			frames->GetFrame( pixels, audio_buffers, frame_number );
			image->PreGetFrame( );
			image->GetFrame( pixels, frame_number );
			sound->GetFrame( audio_buffers, frame_number );

			dv_encode_full_frame(encoder, dv_pixels, e_dv_color_rgb, dv_buffer);
			dv_encode_full_audio(encoder, audio_buffers, info->channels, info->frequency, dv_buffer);
			dv_encode_metadata( dv_buffer, encoder->isPAL, encoder->is16x9, &datetime, frameNum );
			dv_encode_timecode( dv_buffer, encoder->isPAL, frameNum++ );

			if ( fwrite(dv_buffer, (info->isPAL ? 144000 : 120000), 1, output) != 1 )
				throw "Unable to write video - disk full?";

			if (info->end != info->begin)
				UpdateProgress( (gfloat) (( i - info->begin ) ) / (gfloat) (info->end - info->begin) );
		}
		UpdateProgress( (gfloat) 1 );
		
		frames->Close();
		image->Close();
		sound->Close();

		fclose( output );
		output = NULL;

		PlayList temp;
		temp.LoadMediaObject( dvfile );
		common->getPlayList()->InsertPlayList( temp, info->begin );
		common->getPageEditor()->snapshot();
		common->getPageEditor()->ResetBar();
		common->showScenes();
	}
	catch( const char *exc ) 
	{
		modal_message( (char *)exc );
	}

	buttonMutex = true;
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_preview" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_start" ) ), 0 );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_stop" ) ), 1 );
	buttonMutex = false;

	rendering = false;

	if ( output )
		fclose( output );

	dv_encoder_free(encoder);

	delete info;
	delete dv_buffer;
	delete pixels;
	for ( int n = 0; n < 4; n++ )
		delete audio_buffers[n];
}

/** Stop whatever is running (either preview or rendering).
*/

void PageMagick::Stop() 
{
	if ( rendering || previewing ) 
	{
		rendering = false;
		previewing = false;

		buttonMutex = true;
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_preview" ) ), 0 );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_start" ) ), 0 );
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( lookup_widget( window, "togglebutton_magick_stop" ) ), 1 );
		buttonMutex = false;
	}
}

/** Stop the preview.
*/

void PageMagick::StopPreview() 
{
	if ( previewing )
		Stop();
}

/** Update the progress bar for the export process
*/

void PageMagick::UpdateProgress( gfloat val ) 
{
    struct timeval tv;
    long long now;
	static long long lastUpdateTime = 0;

    gettimeofday(&tv, 0);
    now = 1000000 * (long long) tv.tv_sec + (long long) tv.tv_usec;

    /* update every 0.3 second */

    if (val < 0.0) val = 0;
    if (val > 1.0) val = 1.0;

    if ( now > lastUpdateTime + 300000 || val == 1.0) 
	{
        gtk_progress_bar_update( progressBar, val );
        lastUpdateTime = now;
    }

    while (gtk_events_pending()) 
        gtk_main_iteration();
}

/** put the scene begin and end frame numbers into spinners

    \param i the numerical index position of the scene in the playlist1
*/
void PageMagick::selectScene( int i ) 
{
	int begin = 0;
	int end = 0;
	vector <int> scene = common->getPageEditor()->GetScene();

	begin = i == 0 ? 0 : scene[ i - 1 ];
	end = scene[i];	

	GtkNotebook *notebook = GTK_NOTEBOOK( lookup_widget( this->window, "notebook_magick" ) );
	int page = gtk_notebook_get_current_page( notebook );

	if ( page == 0 ) 
	{
		notebook = GTK_NOTEBOOK( lookup_widget( this->window, "notebook_magick_frames" ) );
		page = gtk_notebook_get_current_page( notebook );

		if ( page == 0 ) 
		{
			GtkToggleButton *rangeToggle = GTK_TOGGLE_BUTTON( lookup_widget( window, "radiobutton_magick_range" ) );
			GtkEntry *startSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_magick_start" ) );
			GtkEntry *endSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_magick_end" ) );	
			GtkAdjustment *adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( startSpin ) );
	
			gtk_toggle_button_set_active( rangeToggle, TRUE );
			adjust->lower = 1;
			adjust->upper = scene[ scene.size() - 1 ] + 1;
			gtk_spin_button_set_value( GTK_SPIN_BUTTON( startSpin ), begin + 1);
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

			adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( endSpin ) );
			adjust->lower = 0;
			adjust->upper = scene[ scene.size() - 1 ] + 1;
			gtk_spin_button_set_value( GTK_SPIN_BUTTON( endSpin ), end );
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );
		}
		else if ( page == 1 ) 
		{
			GtkToggleButton *rangeToggle = GTK_TOGGLE_BUTTON( lookup_widget( window, "radiobutton_magick_create_before" ) );
			GtkEntry *startSpin = GTK_ENTRY( lookup_widget( window, "spinbutton_magick_create_before" ) );
			GtkAdjustment *adjust = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON( startSpin ) );
			gtk_toggle_button_set_active( rangeToggle, TRUE );
			adjust->lower = 1;
			adjust->upper = scene[ scene.size() - 1 ] + 1;
			gtk_spin_button_set_value( GTK_SPIN_BUTTON( startSpin ), begin + 1);
			gtk_signal_emit_by_name( GTK_OBJECT( adjust ), "changed" );

		}
	}

	RefreshStatus();
}

/** Refresh the status label to show the current selections.
*/

void PageMagick::RefreshStatus( )
{
	char temp[ 10240 ] = "";
	GtkNotebook *notebookFrame = GTK_NOTEBOOK( lookup_widget( this->window, "notebook_magick_frames" ) );
	GtkNotebook *notebookImage = GTK_NOTEBOOK( lookup_widget( this->window, "notebook_magick_video" ) );
	GtkNotebook *notebookAudio = GTK_NOTEBOOK( lookup_widget( this->window, "notebook_magick_audio" ) );

	int page = gtk_notebook_get_current_page( notebookFrame );
	if ( page == 0 )
		strcpy( temp, "Rewriting frames\n" );
	else if ( page == 1 )
		strcpy( temp, "Creating frames\n" );
	else
		strcpy( temp, "Unknown frame source\n" );

	page = gtk_notebook_get_current_page( notebookImage );

	if ( page == 0 && GetImageFilter() != NULL ) 
	{
		strcat( temp, "with the '" );
		strcat( temp, GetImageFilter()->GetDescription() );
		strcat( temp, "' image filter\n" );
	}
	else if ( page == 1 && GetImageTransition() != NULL )
	{
		strcat( temp, "with the '" );
		strcat( temp, GetImageTransition()->GetDescription() );
		strcat( temp, "' image transition\n" );
	}
	else
		strcpy( temp, "with unknown image manipulation\n" );

	page = gtk_notebook_get_current_page( notebookAudio );
	if ( page == 0 )
	{
		strcat( temp, "and the '" );
		strcat( temp, GetAudioFilter()->GetDescription() );
		strcat( temp, "' audio filter." );
	}
	else if ( page == 1 )
	{
		strcat( temp, "and the '" );
		strcat( temp, GetAudioTransition()->GetDescription() );
		strcat( temp, "' audio transition." );
	}
	else
		strcpy( temp, "and unknown audio manipulation." );

	if ( strcmp( temp, status ) ) 
	{
		strcpy( status, temp );
		GtkLabel *label = GTK_LABEL( lookup_widget( this->window, "label_magick" ) );
		gtk_label_set_text( label, status );
	}
}

/** Return the currently selected image filter.
*/

GDKImageFilter *PageMagick::GetImageFilter( ) const 
{
	return image_filters.Get( );
}

/** Return the currently selected image transition.
*/

GDKImageTransition *PageMagick::GetImageTransition( ) const 
{
	return image_transitions.Get( );
}

/** Return the currently selected audio filter.
*/

GDKAudioFilter *PageMagick::GetAudioFilter( ) const 
{
	return audio_filters.Get( );
}

/** Return the currently selected audio transition.
*/

GDKAudioTransition *PageMagick::GetAudioTransition( ) const 
{
	return audio_transitions.Get( );
}

/** Return the currently selected image creator.
*/

GDKImageCreate *PageMagick::GetImageCreate( ) const 
{
	return image_creators.Get( );
}

/** Load the splash screen for the preview.
*/

void PageMagick::LoadSplash( )
{
	GtkDrawingArea *area = GTK_DRAWING_AREA( lookup_widget( window, "drawingarea_magick_preview" ) );
	common->loadSplash( area );
}
